/* gnubiff -- a mail notification program
 * Copyright (C) 2000-2003 Nicolas Rougier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this software; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
#include <string>
#include <sstream>
#include <glib.h>
#include <sys/stat.h>
#include <utime.h>
#include "Imap4.h"

// ===================================================================
// = Some useful macros (as always) ==================================
// ===================================================================
#ifdef CHECK_SOCKET
#  undef CHECK_SOCKET
#endif
#define CHECK_SOCKET	if (_socket_status == SOCKET_STATUS_ERROR) {socket_close(); return;}



// ===================================================================
// = Static attributes ===============================================
// ===================================================================


// ===================================================================
// = Callbacks "C" binding ===========================================
// ===================================================================


// ===================================================================
// = Imap4 ===========================================================
/*!
     Create a new Imap4 object.
 */
// ===================================================================
Imap4::Imap4 (Biff *owner) : Mailbox (owner) {
  _saved.clear();
}


// ===================================================================
// = ~Imap4 ==========================================================
/*!
     Destroy the Imap4 object.
 */
// ===================================================================
Imap4::~Imap4 (void) {
}

// ===================================================================
// = connect =========================================================
/**
   
 */
// ===================================================================
void Imap4::connect (void) {
  string line;

  // Connection
  socket_open (_owner->_address, _owner->_port);	CHECK_SOCKET;
  socket_read (line);  					CHECK_SOCKET;

  // LOGIN
  line = "A001 LOGIN " + _owner->_user + string (" ") + _owner->_password + string ("\r\n");
  socket_write (line);	CHECK_SOCKET;
  socket_read (line);	CHECK_SOCKET;
  if (line.find ("A001 OK") != 0) {
    _socket_status = SOCKET_STATUS_ERROR;
    return;
  }

  // SELECT
  socket_write ("A002 SELECT INBOX\r\n");
  CHECK_SOCKET;
  gboolean check = FALSE;
  while (socket_read (line)) {
    if (line.find ("A002 OK") != string::npos)
      check = true;
  };
  if (check)
    _socket_status = SOCKET_STATUS_OK;

  return;
}


// ===================================================================
// = update ==========================================================
/**
 */
// ===================================================================
void Imap4::update (void) {
  string line;
  vector<int> buffer;
  
  // By default we consider to an error status
  _status = MAILBOX_ERROR;

  // Check password is not empty
  if (_owner->_password.empty())
    return;

  connect ();
  CHECK_SOCKET;

  // SEARCH NOT SEEN
  socket_write ("A003 SEARCH NOT SEEN\r\n");
  CHECK_SOCKET;
  do {
    socket_read(line);
    CHECK_SOCKET;
  } while (line.find ("* SEARCH") == string::npos);

  // Parse server answer
  line = line.substr (9);
  buffer.clear();
  int n = 0;
  for (guint i=0; i<line.size(); i++) {
    if (line[i] >= '0' && line[i] <= '9')
      n = n*10 + int(line[i]-'0');
    else {
      buffer.push_back (n);
      n = 0;
    }
  }
  if (buffer.empty())
    _status = MAILBOX_EMPTY;
  else if (buffer == _saved)
    _status = MAILBOX_OLD;
  else
    _status = MAILBOX_NEW;

  // We're done
  _saved = buffer;
  //CHECK_SERVER_TAG("A003");

  // Closing connection
  socket_write ("A004 LOGOUT\r\n");
  CHECK_SOCKET;

  socket_close ();
}


// ===================================================================
// = fetch ===========================================================
/**
 */
// ===================================================================
void Imap4::fetch (void) {
  string line;
  vector<int> buffer;
  int saved_status = _status;
  
  // Status will be restored in the end if no problem occured
  _status = MAILBOX_ERROR;

  // Check password is not empty
  if (_owner->_password.empty())
    return;

  // Connection and authentification
  connect ();
  CHECK_SOCKET;


  // SEARCH NOT SEEN
  socket_write ("A003 SEARCH NOT SEEN\r\n");
  CHECK_SOCKET;
  do {
    socket_read (line);
    CHECK_SOCKET;
  } while (line.find ("* SEARCH") == string::npos);

  // Parse server answer
  line = line.substr (9);
  buffer.clear();
  int n = 0;
  for (guint i=0; i<line.size(); i++) {
    if (line[i] >= '0' && line[i] <= '9')
      n = n*10 + int(line[i]-'0');
    else {
      buffer.push_back (n);
      n = 0;
    }
  }

  while (socket_read (line));
  //CHECK_SERVER_TAG("A003");

  // FETCH NOT SEEN
  _unreads.clear();
  vector<string> mail;
  for (guint i=0; (i<buffer.size() && (_unreads.size() < _owner->_max_header_display)); i++) {
    stringstream s;
    s << buffer[i];
    mail.clear();

    string line = string ("A004 FETCH ") + s.str() + string (" (FLAGS BODY.PEEK[HEADER.FIELDS (DATE FROM SUBJECT)])\r\n");
    socket_write (line);
    CHECK_SOCKET; 
    socket_read (line);
    CHECK_SOCKET;
    socket_read (line);
    CHECK_SOCKET;
    while (line.find ("A004 OK") == string::npos) {
      mail.push_back (line.substr(0, line.size()-1));
      socket_read (line);
      CHECK_SOCKET;
    }
    // Remove two last lines which are an empty line and a line with a closing parenthesis
    mail.pop_back();
    mail.pop_back();

    line = string ("A005 FETCH ") + s.str() + string (" (FLAGS BODY.PEEK[TEXT])\r\n");
    socket_write (line);
    CHECK_SOCKET; 
    socket_read (line);
    CHECK_SOCKET;
    socket_read (line);
    CHECK_SOCKET;
    guint j = 0;
    while (line.find ("A005 OK") == string::npos) {
      if (j < 13) {
	mail.push_back (line.substr(0, line.size()-1));
	j++;
      }
      socket_read (line);
      CHECK_SOCKET;
    }
    // Remove two last lines which are an empty line and a line with a closing parenthesis
    mail.pop_back();
    mail.pop_back();
    parse (mail, MAIL_UNREAD);
  }


  // Closing connection
  socket_write ("A006 LOGOUT\r\n");
  CHECK_SOCKET;
  socket_read (line);
  CHECK_SOCKET;

  // We restore status
  _status = saved_status;

  socket_close ();
}
