/* gnubiff -- a mail notification program
 * Copyright (C) 2000-2003 Nicolas Rougier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this software; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

// ===================================================================
/**
   @class Mailbox
   @brief Interface class for a generic mailbox

   This class represents a generic mailbox that will be later
   specialized into File, MH, Maildir, Pop3, Apop or Imap4.
*/
// ===================================================================

#ifndef _MAILBOX_H
#define _MAILBOX_H

#ifdef HAVE_CONFIG_H
#   include "../config/config.h"
#endif
#include <vector>

#ifdef HAVE_LIBSSL
#  include <openssl/ssl.h>
#  include <openssl/err.h>
#endif

#include "Biff.h"


// ===================================================================
// - Constant definitions --------------------------------------------
// ===================================================================
#define MAILBOX_UNITIALIZED	0
#define MAILBOX_ERROR		1
#define MAILBOX_EMPTY		2
#define MAILBOX_OLD		3
#define MAILBOX_NEW		4

#define MAIL_UNREAD		0
#define MAIL_READ		1

#define SOCKET_STATUS_OK	0
#define SOCKET_STATUS_ERROR	1

#define SD_CLOSE		-1

// ===================================================================
// - Type definitions ------------------------------------------------
// ===================================================================


using namespace std;
class Mailbox {
  friend class Popup;

  // ===================================================================
  // - Public methods --------------------------------------------------
  // ===================================================================
public:
  //! @name Base functions
  Mailbox (class Biff *owner);
  virtual ~Mailbox (void);
  //@}

  //! @name I/O functions
  /**
     Get the status of the mailbox (_status):
     - MAILBOX_ERROR (cannot access mailbox),
     - MAILBOX_EMPTY (no mail in mailbox),
     - MAILBOX_OLD (only old mails in mailbox),
     - MAILBOX_NEW (some new mails in mailbox)
  */
  virtual void update (void) = 0;
  /** Fetch unread mails from mailbox */
  virtual void fetch (void) = 0;
  // @}

  //! @name Object manipulations
  void parse (vector<string> &mail, int status = -1);
  // @}

#ifndef DOXYGEN_SHOULD_SKIP_THIS
  //! @name Features consultations / modifications
  gint status (void)			{return _status;}
  guint unreads (void)			{return _unreads.size();}
  // @}
#endif


  // ===================================================================
  // - Public attributes -----------------------------------------------
  // ===================================================================
public:



  // ===================================================================
  // - Protected methods -----------------------------------------------
  // ===================================================================
protected:
  void socket_open  (string &hostname, unsigned short int port);
  void socket_close (void);
  void socket_write (string line);
  gint socket_read  (string &line);

  // ===================================================================
  // - Protected attributes --------------------------------------------
  // ===================================================================
protected:
  class Biff *		_owner;		/*!< \brief Owner of this interface */
  gint 			_status;	/*!< \brief Status of the mailbox */
  vector<header>	_unreads;	/*!< \brief Collected unread mail */
#ifdef HAVE_LIBSSL
  static SSL_CTX *	_context;	/*!< \brief OpenSSL context object */
  SSL *			_ssl;		/*!< \brief OpenSSL socket object */
#endif
  int			_sd;		/*!< \brief Regular socket descriptor */
  int			_socket_status;	/*!< \brief Socket status */


  // ===================================================================
  // - Private methods -------------------------------------------------
  // ===================================================================
 private:

  // ===================================================================
  // - Private attributes ----------------------------------------------
  // ===================================================================
private:
};

#endif /* _MAILBOX_H */
